#include "plugin.hpp"


struct LowFrequencyOscillator {
	float phase = 0.f;
	float pw = 0.5f;
	float freq = 1.f;
	bool offset = false;
	bool invert = false;
	dsp::SchmittTrigger resetTrigger;

	LowFrequencyOscillator() {}
	void setPitch(float pitch) {
		pitch = std::fmin(pitch, 10.f);
		freq = std::pow(2.f, pitch);
	}
	void setPulseWidth(float pw_) {
		const float pwMin = 0.01f;
		pw = clamp(pw_, pwMin, 1.f - pwMin);
	}
	void setReset(float reset) {
		if (resetTrigger.process(reset / 0.01f)) {
			phase = 0.f;
		}
	}
	void step(float dt) {
		float deltaPhase = std::fmin(freq * dt, 0.5f);
		phase += deltaPhase;
		if (phase >= 1.f)
			phase -= 1.f;
	}
	float sin() {
		if (offset)
			return 1.f - std::cos(2*M_PI * phase) * (invert ? -1.f : 1.f);
		else
			return std::sin(2*M_PI * phase) * (invert ? -1.f : 1.f);
	}
	float tri(float x) {
		return 4.f * std::fabs(x - std::round(x));
	}
	float tri() {
		if (offset)
			return tri(invert ? phase - 0.5f : phase);
		else
			return -1.f + tri(invert ? phase - 0.25f : phase - 0.75f);
	}
	float saw(float x) {
		return 2.f * (x - std::round(x));
	}
	float saw() {
		if (offset)
			return invert ? 2.f * (1.f - phase) : 2.f * phase;
		else
			return saw(phase) * (invert ? -1.f : 1.f);
	}
	float sqr() {
		float sqr = (phase < pw) ^ invert ? 1.f : -1.f;
		return offset ? sqr + 1.f : sqr;
	}
	float light() {
		return std::sin(2*M_PI * phase);
	}
};


struct LFO : Module {
	enum ParamIds {
		OFFSET_PARAM,
		INVERT_PARAM,
		FREQ_PARAM,
		FM1_PARAM,
		FM2_PARAM,
		PW_PARAM,
		PWM_PARAM,
		NUM_PARAMS
	};
	enum InputIds {
		FM1_INPUT,
		FM2_INPUT,
		RESET_INPUT,
		PW_INPUT,
		NUM_INPUTS
	};
	enum OutputIds {
		SIN_OUTPUT,
		TRI_OUTPUT,
		SAW_OUTPUT,
		SQR_OUTPUT,
		NUM_OUTPUTS
	};
	enum LightIds {
		PHASE_POS_LIGHT,
		PHASE_NEG_LIGHT,
		NUM_LIGHTS
	};

	LowFrequencyOscillator oscillator;

	LFO() {
		config(NUM_PARAMS, NUM_INPUTS, NUM_OUTPUTS, NUM_LIGHTS);
		configParam(OFFSET_PARAM, 0.f, 1.f, 1.f);
		configParam(INVERT_PARAM, 0.f, 1.f, 1.f);
		configParam(FREQ_PARAM, -8.f, 10.f, 1.f);
		configParam(FM1_PARAM, 0.f, 1.f, 0.f);
		configParam(PW_PARAM, 0.f, 1.f, 0.5f);
		configParam(FM2_PARAM, 0.f, 1.f, 0.f);
		configParam(PWM_PARAM, 0.f, 1.f, 0.f);
	}

	void process(const ProcessArgs &args) override {
		oscillator.setPitch(params[FREQ_PARAM].getValue() + params[FM1_PARAM].getValue() * inputs[FM1_INPUT].getVoltage() + params[FM2_PARAM].getValue() * inputs[FM2_INPUT].getVoltage());
		oscillator.setPulseWidth(params[PW_PARAM].getValue() + params[PWM_PARAM].getValue() * inputs[PW_INPUT].getVoltage() / 10.f);
		oscillator.offset = (params[OFFSET_PARAM].getValue() > 0.f);
		oscillator.invert = (params[INVERT_PARAM].getValue() <= 0.f);
		oscillator.step(args.sampleTime);
		oscillator.setReset(inputs[RESET_INPUT].getVoltage());

		outputs[SIN_OUTPUT].setVoltage(5.f * oscillator.sin());
		outputs[TRI_OUTPUT].setVoltage(5.f * oscillator.tri());
		outputs[SAW_OUTPUT].setVoltage(5.f * oscillator.saw());
		outputs[SQR_OUTPUT].setVoltage(5.f * oscillator.sqr());

		lights[PHASE_POS_LIGHT].setSmoothBrightness(oscillator.light(), args.sampleTime);
		lights[PHASE_NEG_LIGHT].setSmoothBrightness(-oscillator.light(), args.sampleTime);
	}

};



struct LFOWidget : ModuleWidget {
	LFOWidget(LFO *module) {
		setModule(module);
		setPanel(APP->window->loadSvg(asset::plugin(pluginInstance, "res/LFO-1.svg")));

		addChild(createWidget<ScrewSilver>(Vec(15, 0)));
		addChild(createWidget<ScrewSilver>(Vec(box.size.x-30, 0)));
		addChild(createWidget<ScrewSilver>(Vec(15, 365)));
		addChild(createWidget<ScrewSilver>(Vec(box.size.x-30, 365)));

		addParam(createParam<CKSS>(Vec(15, 77), module, LFO::OFFSET_PARAM));
		addParam(createParam<CKSS>(Vec(119, 77), module, LFO::INVERT_PARAM));

		addParam(createParam<RoundHugeBlackKnob>(Vec(47, 61), module, LFO::FREQ_PARAM));
		addParam(createParam<RoundLargeBlackKnob>(Vec(23, 143), module, LFO::FM1_PARAM));
		addParam(createParam<RoundLargeBlackKnob>(Vec(91, 143), module, LFO::PW_PARAM));
		addParam(createParam<RoundLargeBlackKnob>(Vec(23, 208), module, LFO::FM2_PARAM));
		addParam(createParam<RoundLargeBlackKnob>(Vec(91, 208), module, LFO::PWM_PARAM));

		addInput(createInput<PJ301MPort>(Vec(11, 276), module, LFO::FM1_INPUT));
		addInput(createInput<PJ301MPort>(Vec(45, 276), module, LFO::FM2_INPUT));
		addInput(createInput<PJ301MPort>(Vec(80, 276), module, LFO::RESET_INPUT));
		addInput(createInput<PJ301MPort>(Vec(114, 276), module, LFO::PW_INPUT));

		addOutput(createOutput<PJ301MPort>(Vec(11, 320), module, LFO::SIN_OUTPUT));
		addOutput(createOutput<PJ301MPort>(Vec(45, 320), module, LFO::TRI_OUTPUT));
		addOutput(createOutput<PJ301MPort>(Vec(80, 320), module, LFO::SAW_OUTPUT));
		addOutput(createOutput<PJ301MPort>(Vec(114, 320), module, LFO::SQR_OUTPUT));

		addChild(createLight<SmallLight<GreenRedLight>>(Vec(99, 42.5f), module, LFO::PHASE_POS_LIGHT));
	}
};


Model *modelLFO = createModel<LFO, LFOWidget>("LFO");


struct LFO2 : Module {
	enum ParamIds {
		OFFSET_PARAM,
		INVERT_PARAM,
		FREQ_PARAM,
		WAVE_PARAM,
		FM_PARAM,
		NUM_PARAMS
	};
	enum InputIds {
		FM_INPUT,
		RESET_INPUT,
		WAVE_INPUT,
		NUM_INPUTS
	};
	enum OutputIds {
		INTERP_OUTPUT,
		NUM_OUTPUTS
	};
	enum LightIds {
		PHASE_POS_LIGHT,
		PHASE_NEG_LIGHT,
		NUM_LIGHTS
	};

	LowFrequencyOscillator oscillator;

	LFO2() {
		config(NUM_PARAMS, NUM_INPUTS, NUM_OUTPUTS, NUM_LIGHTS);
		configParam(OFFSET_PARAM, 0.f, 1.f, 1.f);
		configParam(INVERT_PARAM, 0.f, 1.f, 1.f);
		configParam(FREQ_PARAM, -8.f, 10.f, 1.f);
		configParam(WAVE_PARAM, 0.f, 3.f, 1.5f);
		configParam(FM_PARAM, 0.f, 1.f, 0.5f);
	}

	void process(const ProcessArgs &args) override {
		float deltaTime = args.sampleTime;
		oscillator.setPitch(params[FREQ_PARAM].getValue() + params[FM_PARAM].getValue() * inputs[FM_INPUT].getVoltage());
		oscillator.offset = (params[OFFSET_PARAM].getValue() > 0.f);
		oscillator.invert = (params[INVERT_PARAM].getValue() <= 0.f);
		oscillator.step(deltaTime);
		oscillator.setReset(inputs[RESET_INPUT].getVoltage());

		float wave = params[WAVE_PARAM].getValue() + inputs[WAVE_INPUT].getVoltage();
		wave = clamp(wave, 0.f, 3.f);
		float interp;
		if (wave < 1.f)
			interp = crossfade(oscillator.sin(), oscillator.tri(), wave);
		else if (wave < 2.f)
			interp = crossfade(oscillator.tri(), oscillator.saw(), wave - 1.f);
		else
			interp = crossfade(oscillator.saw(), oscillator.sqr(), wave - 2.f);
		outputs[INTERP_OUTPUT].setVoltage(5.f * interp);

		lights[PHASE_POS_LIGHT].setSmoothBrightness(oscillator.light(), deltaTime);
		lights[PHASE_NEG_LIGHT].setSmoothBrightness(-oscillator.light(), deltaTime);
	}
};


struct LFO2Widget : ModuleWidget {
	LFO2Widget(LFO2 *module) {
		setModule(module);
		setPanel(APP->window->loadSvg(asset::plugin(pluginInstance, "res/LFO-2.svg")));

		addChild(createWidget<ScrewSilver>(Vec(15, 0)));
		addChild(createWidget<ScrewSilver>(Vec(box.size.x-30, 0)));
		addChild(createWidget<ScrewSilver>(Vec(15, 365)));
		addChild(createWidget<ScrewSilver>(Vec(box.size.x-30, 365)));

		addParam(createParam<CKSS>(Vec(62, 150), module, LFO2::OFFSET_PARAM));
		addParam(createParam<CKSS>(Vec(62, 215), module, LFO2::INVERT_PARAM));

		addParam(createParam<RoundHugeBlackKnob>(Vec(18, 60), module, LFO2::FREQ_PARAM));
		addParam(createParam<RoundLargeBlackKnob>(Vec(11, 142), module, LFO2::WAVE_PARAM));
		addParam(createParam<RoundLargeBlackKnob>(Vec(11, 207), module, LFO2::FM_PARAM));

		addInput(createInput<PJ301MPort>(Vec(11, 276), module, LFO2::FM_INPUT));
		addInput(createInput<PJ301MPort>(Vec(54, 276), module, LFO2::RESET_INPUT));
		addInput(createInput<PJ301MPort>(Vec(11, 319), module, LFO2::WAVE_INPUT));

		addOutput(createOutput<PJ301MPort>(Vec(54, 319), module, LFO2::INTERP_OUTPUT));

		addChild(createLight<SmallLight<GreenRedLight>>(Vec(68, 42.5f), module, LFO2::PHASE_POS_LIGHT));
	}
};


Model *modelLFO2 = createModel<LFO2, LFO2Widget>("LFO2");
